@@ -18,71 +18,92 @@ def generate_terrain(x_range: tuple = (0, 500),
                      zfactor: int = 4000,
                      full_extent: Optional[str] = None) -> xr.DataArray:
     """
-    Generates a pseudo-random terrain which can be helpful
-    for testing raster functions
+    Generates a pseudo-random terrain which can be helpful for testing
+    raster functions.
 
-    Parameters:
+    Parameters
     ----------
-    x_range: tuple (default = (0, 500))
+    x_range : tuple, default=(0, 500)
         Range of x values.
-    x_range: tuple (default = (0, 500))
+    x_range : tuple, default=(0, 500)
         Range of y values.
-    width: int (default = 25)
+    width : int, default=25
         Width of output data array in pixels.
-    height: int (default = 30)
+    height : int, default=30
         Height of output data array in pixels.
-    canvas: ds.Canvas (default = None)
-        Instance for passing output dimensions / ranges
-    seed: int (default = 10)
+    canvas : ds.Canvas, default=None
+        Instance for passing output dimensions / ranges.
+    seed : int, default=10
         Seed for random number generator.
-    zfactor: int (default = 4000)
+    zfactor : int, default=4000
         Multipler for z values.
-    full_extent: str, optional (default = None)
+    full_extent : str, default=None
         bbox<xmin, ymin, xmax, ymax>. Full extent of coordinate system.
 
-    Returns:
+    Returns
+    -------
+    terrain : xr.DataArray
+        2D array of generated terrain values.
+
+    References
     ----------
-    terrain: xarray.DataArray
-        2D array of generated terrain.
+        - Michael McHugh: https://www.youtube.com/watch?v=O33YV4ooHSo
+        - Red Blob Games: https://www.redblobgames.com/maps/terrain-from-noise/
 
-    Notes:
-    ----------
-    Algorithm References:
-    - This was inspired by Michael McHugh's 2016 PyCon Canada talk: # noqa
-        - Michael McHugh, PyCon Canada, YouTube, Dec. 9, 2016, https://www.youtube.com/watch?v=O33YV4ooHSo, Accessed Apr. 22, 2021. # noqa
-    - Patel, A, Red Blob Games, Making maps with noise functions, https://www.redblobgames.com/maps/terrain-from-noise/, Accessed Apr. 22, 2021. # noqa
+    Examples
+    --------
+    .. plot::
+       :include-source:
 
-    Examples:
-    ----------
-    Imports
-    >>> import datashader as ds
-    >>> from datashader.transfer_functions import shade
-    >>> from xrspatial import generate_terrain
+        import datashader as ds
+        import matplotlib.pyplot as plt
+        from xrspatial import generate_terrain, aspect
 
-    Create Canvas
-    >>> cvs = ds.Canvas(plot_width=800,
-    >>>                 plot_height=600,
-    >>>                 x_range=(-20e6, 20e6),
-    >>>                 y_range=(-20e6, 20e6))
+        # Create Canvas
+        W = 500
+        H = 300
+        cvs = ds.Canvas(plot_width = W,
+                        plot_height = H,
+                        x_range = (-20e6, 20e6),
+                        y_range = (-20e6, 20e6))
 
-    Generate Terrain Data Array
-    >>> terrain = generate_terrain(canvas = cvs)
-    >>> print(terrain)
-    <xarray.DataArray 'terrain' (y: 600, x: 800)>
-    array([[0., 0., 0., ..., 0., 0., 0.],
-           [0., 0., 0., ..., 0., 0., 0.],
-           [0., 0., 0., ..., 0., 0., 0.],
-           ...,
-           [0., 0., 0., ..., 0., 0., 0.],
-           [0., 0., 0., ..., 0., 0., 0.],
-           [0., 0., 0., ..., 0., 0., 0.]])
-    Coordinates:
-      * x (x) float64 -1.998e+07 -1.992e+07 ... 1.992e+07 1.997e+07
-      * y (y) float64 -1.997e+07 -1.99e+07 -1.983e+07 ... 1.99e+07 1.997e+07
-    Attributes:
-        res: 1
+        # Generate Example Terrain
+        terrain_agg = generate_terrain(canvas = cvs)
+
+        # Edit Attributes
+        terrain_agg = terrain_agg.assign_attrs(
+            {
+                'Description': 'Example Terrain',
+                'units': 'km',
+                'Max Elevation': '4000',
+            }
+        )
+
+        terrain_agg = terrain_agg.rename({'x': 'lon', 'y': 'lat'})
+        terrain_agg = terrain_agg.rename('Elevation')
+
+        # Plot Terrain
+        terrain_agg.plot(cmap = 'terrain', aspect = 2, size = 4)
+        plt.title("Terrain")
+        plt.ylabel("latitude")
+        plt.xlabel("longitude")
+
+    .. sourcecode:: python
+
+        >>> print(terrain_agg[200:203, 200:202])
+        <xarray.DataArray 'Elevation' (lat: 3, lon: 2)>
+        array([[1264.02249454, 1261.94748873],
+               [1285.37061171, 1282.48046696],
+               [1306.02305679, 1303.40657515]])
+        Coordinates:
+          * lon      (lon) float64 -3.96e+06 -3.88e+06
+          * lat      (lat) float64 6.733e+06 6.867e+06 7e+06
+        Attributes:
+            res:            1
+            Description:    Example Terrain
+            units:          km
+            Max Elevation:  4000
     """
-
     def _gen_heights(bumps):
         out = np.zeros(len(bumps))
         for i, b in enumerate(bumps):
